import pandas as pd
from folium.plugins import HeatMap, HeatMapWithTime
import folium
from .geo import *
import numpy as np
import branca.colormap as cm
from branca.element import Template, MacroElement

colors_price = [
    'green',
    'blue',
    'red',
    'yellow'
]
colors_flood = ['lightgreen','green','darkgreen','beige', 'blue','orange','pink',
        'lightred', 'red','darkred',"purple"]

def colorCall_price(val):
  val = int(val)
  return colors_price[val]

def colorCall_flood(val):
  val = int(val)
  return colors_flood[val]

def func_rainfall_class(x):
    if x < 2:
        return 0
    elif x < 4:
        return 1
    elif x < 50:
        return 2
    return 3


class Preprocess:
    def __init__(self,price="flood_tool/resources/predicted_data.csv",
    risk_label="flood_tool/resources/prediction_riskLabel.csv",
    station_data="flood_tool/resources/stations.csv",
    annual_risk = "flood_tool/resources/AnnualRiskPredict.csv"):

        self.house_price = pd.read_csv(price)
        self.risk_label = pd.read_csv(risk_label)
        self.station_data = pd.read_csv(station_data)
        self.annual_risk = pd.read_csv(annual_risk)

    def merge(self, filename):
        
        """Merge the given data with a station to get location information.
        
        Parameters
        ----------

        filename: path to datafram to be read in via read_csv."""
        
        data = pd.read_csv(filename)
        data = pd.merge(data, self.station_data, how="left", on="stationReference")
        data = data.dropna(subset=["latitude", "longitude"])
        data["value"] = data["value"].apply(lambda x: x if isinstance(x, str) else str(x))
        index = data["value"].apply(lambda x: False if '|' in x else True)
        data = data[index]
        data["value"] = data["value"].apply(lambda x: float(x))

        rainfall = data[data["parameter"] == "rainfall"]
        river = data[data["parameter"] == "level"]
        return rainfall, river

    def get_class(self,filename):
        
        """Calculate the class (high or low) for rainfall and river data using 
        percentages and typical high and low range data for each station.
        
        Parameters
        ----------

        filename: path to datafram to be read in via read_csv."""
        
        rainfall,river = self.merge(filename)
        rainfall = rainfall.groupby("stationReference")["value"].mean() * 4
        rainfall_df = rainfall.to_frame()
        rainfall_df = rainfall_df.reset_index()
        rainfall_df["class"] = rainfall_df["value"].apply(lambda x: func_rainfall_class(x))
        rainfall_df = pd.merge(rainfall_df, self.station_data, how="left", on="stationReference")

        river_result = river.groupby("stationReference")["value"].mean()
        river_result_df = river_result.to_frame()
        river_result_df = river_result_df.reset_index()
        river_result_df = pd.merge(river_result_df, self.station_data, how="left", on="stationReference")
        vs = river_result_df["value"].tolist()
        highs = river_result_df["typicalRangeHigh"].tolist()
        lows = river_result_df["typicalRangeLow"].tolist()
        class_list = []
        percentage_list = []
        for v, h, l in zip(vs, highs, lows):
            if v > h:
                class_list.append(2)
            elif v < l:
                class_list.append(0)
            else:
                class_list.append(1)
            percentage_list.append((v - l) / (h - l))
        river_result_df["class"] = class_list
        river_result_df["percentage"] = percentage_list
        return rainfall_df, river_result_df
   

    def house_price_preprocess(self):
        
        """Preprocess the house price data to get it ready for visualization 
        through defining high, medium, and low prices."""
        
        coord = []
        for i in range(len(self.house_price)):
            coord.append(get_gps_lat_long_from_easting_northing(self.house_price["easting"][i],
                                                                self.house_price["northing"][i]))
        self.house_price["lat"] = [x[0][0] for x in coord]
        self.house_price["long"] = [x[1][0] for x in coord]
        price = self.house_price[['lat', 'long', 'predicted_MedianPrice']]

        def func(x):
            if x < np.percentile(price["predicted_MedianPrice"],25):
                return 0
            elif x < np.percentile(price["predicted_MedianPrice"],50):
                return 1
            elif np.percentile(price["predicted_MedianPrice"],75):
                return 2
            return 3

        price["class"] = price["predicted_MedianPrice"].apply(lambda x: func(x))
        return price
    
    def risk_label_preprocess(self):
        
        """Preprocess the risk label data by converting easting/northing to latitude/longitude."""
        
        coord = []
        for i in range(len(self.risk_label)):
            coord.append(get_gps_lat_long_from_easting_northing(self.risk_label["easting"][i],
                                                                self.risk_label["northing"][i]))
        self.risk_label["lat"] = [x[0][0] for x in coord]
        self.risk_label["long"] = [x[1][0] for x in coord]
        risk_predictions = self.risk_label[['lat', 'long', 'riskLabel']]

        return risk_predictions
    
    def Annual_risk_preprocess(self):
        coord = []
        for i in range(len(self.annual_risk)):
            coord.append(get_gps_lat_long_from_easting_northing(self.annual_risk["easting"][i],
                                                                self.annual_risk["northing"][i]))


        self.annual_risk["lat"] = [x[0][0] for x in coord]
        self.annual_risk["long"] = [x[1][0] for x in coord]
        risk_predictions = self.annual_risk[['lat', 'long', 'AnnualRisk']]
        risk_predictions["AnnualRisk"] = np.log(risk_predictions["AnnualRisk"])

        def func(x):
            if x < 2.2:
                return 0
            elif x < 6:
                return 1
            else:return 2

        risk_predictions["class"] = risk_predictions["AnnualRisk"].apply(lambda x: func(x))
        return risk_predictions


    def zip_data(self, data_df, key="value"):
        
        """Zip the latitude/longitude and value, which has the key name of parameter key.
        
        Parameters
        ----------

        data_df: path to file to read in through pd.read_csv."""
        
        lat = data_df["latitude"].tolist()
        lon = data_df["longitude"].tolist()
        v = data_df[key].tolist()
        return list(zip(lat, lon, v))


class Visualize():

    def __init__(self,price="flood_tool/resources/predicted_data.csv",
    risk_label="flood_tool/resources/prediction_riskLabel.csv",
    annual_risk = "flood_tool/resources/AnnualRiskPredict.csv"):
        self.house_price = pd.read_csv(price)
        self.risk_label = pd.read_csv(risk_label)
        self.annual_risk = pd.read_csv(annual_risk)


    def plot_HeatMapWithTime(self, data):
        # using datetime to plot the original value of the inputed data
        date_time = list(data["dateTime"].sort_values().astype("str").unique())
        loc = 'Heat map of Rainfall/Level distribution with time over a wet/typical day'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)
        tdata = []
        for _, d in data.groupby("dateTime"):
            tdata.append([[row["latitude"], row["longitude"], row["value"]] for _, row in d.iterrows()])
        map = folium.Map([51.4, 0.3], zoom_start=7, control_scale=True)
        map.get_root().html.add_child(folium.Element(title_html))
        HeatMapWithTime(tdata, index=date_time, radius=30).add_to(map)
        return map   

    def plot_wet_and_typical_class(self, wet_data_c, ty_data_c, key="value"):
        
        """Create plot of river and rainfall data on wet and typical days.
        
        Parameters
        ----------

        wet_data_c: path to data on a wet day in the United Kingdom, in terms of river level
        ty_data_c: path to data on a typical day in the United Kingdom, in terms of river level"""

        loc = 'Comparing Wet Day and Typical Day distributions'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)
        
        if key == "percentage":
            wet_data_c = wet_data_c.dropna(subset=["percentage"])
            ty_data_c = ty_data_c.dropna(subset=["percentage"])
        
        wet_river_zip = Preprocess.zip_data(self,wet_data_c, key)
        ty_river_zip = Preprocess.zip_data(self,ty_data_c, key)
        map = folium.Map([52.5, 0.], zoom_start=6,zoom_end=15,control_scale=True)
        map.get_root().html.add_child(folium.Element(title_html))
        layer_wet_river = folium.FeatureGroup(name="Wet day")
        layer_typical_river = folium.FeatureGroup(name="Typical day")
        HeatMap(wet_river_zip, radius=15).add_to(layer_wet_river)
        HeatMap(ty_river_zip, radius=15).add_to(layer_typical_river)
        map.add_child(layer_wet_river)
        map.add_child(layer_typical_river)
        map.add_child(folium.map.LayerControl())
        return map

    def plot_heat_map(self, data):
        
        """Create heatmap using datetime to plot the original value of the data.
        
        Parameters
        ----------

        data: path to dataframe."""
        
        date_time = list(data["dateTime"].sort_values().astype("str").unique())
        tdata = []
        for _, d in data.groupby("dateTime"):
            tdata.append([[row["latitude"], row["longitude"], row["value"]] for _, row in d.iterrows()])
        map = folium.Map([53, 0.3], zoom_start=8, control_scale=True)
        HeatMapWithTime(tdata, index=date_time, radius=20).add_to(map)
        return map

    def plot_class_map(self, wet_data_c, ty_data_c, type="rainfall"):
        
        """Create a visualization of the class (defined in get_class) of rainfall on a wet and typical day.
        
        Parameters
        ----------

        wet_data_c: path to data on a wet day in the United Kingdom, in terms of rainfall
        ty_data_c: path to data on a typical day in the United Kingdom, in terms of rainfall"""
        
        wet_zip = Preprocess.zip_data(self,wet_data_c, "class")
        ty_zip = Preprocess.zip_data(self,ty_data_c, "class")

        loc = 'Class distribution for Rainfall and Tidal/River Level (Typical/Wet day)'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)

        map = folium.Map([51.4,0.], zoom_start=7, control_scale=True)
        map.get_root().html.add_child(folium.Element(title_html))
        

        layer_wet = folium.FeatureGroup(name="Wet day")
        layer_typical = folium.FeatureGroup(name="Typical day")
        if type == "rainfall":
            layer_wet = self.plot_rain_class(wet_zip, layer_wet)
            layer_typical = self.plot_rain_class(ty_zip, layer_typical)
        else:
            layer_wet = self.plot_river_class(wet_zip, layer_wet)
            layer_typical = self.plot_river_class(ty_zip, layer_typical)
        map.add_child(layer_wet)
        map.add_child(layer_typical)
        map.add_child(folium.map.LayerControl())
        return map
    
    def plot_rain_class(self, data, map):

        """Create a visualization of the rainfall classes.
        Classes are 
        slight: <2mm per hour,
        moderate: 2-4mm per hour,
        high: 4-50mm per hour,
        violent: >50mm per hour"""
        
        radiuses = [2000, 5000, 8000, 10000]

        colormap = ["green", "yellow", "red", "black"]
        
        template = """
        {% macro html(this, kwargs) %}
        
        <div id='maplegend' class='maplegend' 
            style='position: absolute; z-index:9999; border:2px solid grey; background-color:rgba(255, 255, 255, 0.8);
            border-radius:6px; padding: 10px; font-size:14px; right: 20px; bottom: 20px;'>
            
        <div class='legend-title'>Risk Label Legend</div>
        <ul class='legend-labels'>
            <li><span style='background:green;opacity:0.7;'></span>slight rainfall(<2mm per hr)</li>
            <li><span style='background:blue;opacity:0.7;'></span>moderate rainfall(2mm-4mm per hr)</li>
            <li><span style='background:red;opacity:0.7;'></span> high rainfall(4mm-50mm per hr)</li>
            <li><span style='background:yellow;opacity:0.7;'></span>violent rainfall(>50mm per hr)</li>
            
        </ul>
        </div>
        
        </body>
        </html>

        <style type='text/css'>
        .maplegend .legend-title {
            text-align: left;
            margin-bottom: 5px;
            font-weight: bold;
            font-size: 90%;
            }
        .maplegend .legend-scale ul {
            margin: 0;
            margin-bottom: 5px;
            padding: 0;
            float: left;
            list-style: none;
            }
        .maplegend .legend-scale ul li {
            font-size: 80%;
            list-style: none;
            margin-left: 0;
            line-height: 18px;
            margin-bottom: 2px;
            }
        .maplegend ul.legend-labels li span {
            display: block;
            float: left;
            height: 16px;
            width: 30px;
            margin-right: 5px;
            margin-left: 0;
            border: 1px solid #999;
            }
        .maplegend .legend-source {
            font-size: 80%;
            color: #777;
            clear: both;
            }
        .maplegend a {
            color: #777;
            }
        </style>
        {% endmacro %}"""

        macro = MacroElement()
        macro._template = Template(template)


        map.get_root().add_child(macro)
        for la, lo, c in data:
            folium.Circle(location=(la, lo), radius=radiuses[c], color=colormap[c], popup=c, fill=True,
                          fill_opacity=0.7).add_to(map)
        
        
        return map

    def plot_river_class(self, data, map):

        
        """Create a visualization for the river classes which are 
        low: lower than typical river level, 
        normal: between low and high,
        high: greater than typical river level"""

        radiuses = [3000, 2000, 10000]

        colormap = ["green", "yellow", "red"]
        template = """
        {% macro html(this, kwargs) %}
        
        <div id='maplegend' class='maplegend' 
            style='position: absolute; z-index:9999; border:2px solid grey; background-color:rgba(255, 255, 255, 0.8);
            border-radius:6px; padding: 10px; font-size:14px; right: 20px; bottom: 20px;'>
            
        <div class='legend-title'>Risk Label Legend</div>
        <ul class='legend-labels'>
            <li><span style='background:green;opacity:0.7;'></span>Level lower than typical low</li>
            <li><span style='background:yellow;opacity:0.7;'></span>Normal (between typical high and low)</li>
            <li><span style='background:red;opacity:0.7;'></span> Level higher than typical high</li>
            
            
        </ul>
        </div>
        
        </body>
        </html>

        <style type='text/css'>
        .maplegend .legend-title {
            text-align: left;
            margin-bottom: 5px;
            font-weight: bold;
            font-size: 90%;
            }
        .maplegend .legend-scale ul {
            margin: 0;
            margin-bottom: 5px;
            padding: 0;
            float: left;
            list-style: none;
            }
        .maplegend .legend-scale ul li {
            font-size: 80%;
            list-style: none;
            margin-left: 0;
            line-height: 18px;
            margin-bottom: 2px;
            }
        .maplegend ul.legend-labels li span {
            display: block;
            float: left;
            height: 16px;
            width: 30px;
            margin-right: 5px;
            margin-left: 0;
            border: 1px solid #999;
            }
        .maplegend .legend-source {
            font-size: 80%;
            color: #777;
            clear: both;
            }
        .maplegend a {
            color: #777;
            }
        </style>
        {% endmacro %}"""

        macro = MacroElement()
        macro._template = Template(template)
        map.get_root().add_child(macro)
        

        for la, lo, c in data:
            folium.Circle(location=(la, lo), radius=radiuses[c], color=colormap[c], popup=c, fill=True,
                          fill_opacity=0.7).add_to(map)
        
        return map

    def house_price_map(self):
        
        """Create a circle plot to visualize the median house price in different postcodes of the United Kingdom"""
        
        price = Preprocess.house_price_preprocess(self)
        import branca.colormap as cm
        cs = price["class"]

        data = zip(price["lat"],price["long"],cs)

        loc = 'Predicted Median House Prices'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)

        map = folium.Map([51.4,0.], zoom_start=8,zoom_end =15)

        map.get_root().html.add_child(folium.Element(title_html))

        for la,lo,c in data:
            folium.Circle(location=(la,lo),radius=1000,popup=c,color=False,
            fill_color = colorCall_price(c),fill=True,fill_opacity=0.5).add_to(map)

        template = """
        {% macro html(this, kwargs) %}
        
        <div id='maplegend' class='maplegend' 
            style='position: absolute; z-index:9999; border:2px solid grey; background-color:rgba(255, 255, 255, 0.8);
            border-radius:6px; padding: 10px; font-size:14px; right: 20px; bottom: 20px;'>
            
        <div class='legend-title'>Risk Label Legend</div>
        <ul class='legend-labels'>
            <li><span style='background:green;opacity:0.7;'></span>Very Low prices (0-25%)</li>
            <li><span style='background:blue;opacity:0.7;'></span>Low prices (25%-50%)</li>
            <li><span style='background:red;opacity:0.7;'></span>High prices (50%-75%)</li>
            <li><span style='background:yellow;opacity:0.7;'></span>Ver High prices(above 75%)</li>
            
        </ul>
        </div>
        
        </body>
        </html>

        <style type='text/css'>
        .maplegend .legend-title {
            text-align: left;
            margin-bottom: 5px;
            font-weight: bold;
            font-size: 90%;
            }
        .maplegend .legend-scale ul {
            margin: 0;
            margin-bottom: 5px;
            padding: 0;
            float: left;
            list-style: none;
            }
        .maplegend .legend-scale ul li {
            font-size: 80%;
            list-style: none;
            margin-left: 0;
            line-height: 18px;
            margin-bottom: 2px;
            }
        .maplegend ul.legend-labels li span {
            display: block;
            float: left;
            height: 16px;
            width: 30px;
            margin-right: 5px;
            margin-left: 0;
            border: 1px solid #999;
            }
        .maplegend .legend-source {
            font-size: 80%;
            color: #777;
            clear: both;
            }
        .maplegend a {
            color: #777;
            }
        </style>
        {% endmacro %}"""

        macro = MacroElement()
        macro._template = Template(template)

        map.get_root().add_child(macro)
        return map


    def risk_label_map(self):
        
        """Create a circle plot to visualize the risk label for different postcodes in the United Kingdom.
        Risk label describes how at risk a property is of flooding."""

        risk_predictions = Preprocess.risk_label_preprocess(self)
        cs = risk_predictions["riskLabel"]

        data = zip(risk_predictions["lat"],risk_predictions["long"],cs)

        

        loc = 'Predicted Flood risk'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)

        map = folium.Map([51.3,0.], zoom_start=8,zoom_end =15)

        map.get_root().html.add_child(folium.Element(title_html))


        for la,lo,c in data:
            radius = ((c+1)*10)**2
            folium.Circle(location=(la,lo),radius=radius,color=False,
            fill_color = colorCall_flood(c),fill=True,fill_opacity=0.5).add_to(map)

        
        template = """
        {% macro html(this, kwargs) %}
        
        <div id='maplegend' class='maplegend' 
            style='position: absolute; z-index:9999; border:2px solid grey; background-color:rgba(255, 255, 255, 0.8);
            border-radius:6px; padding: 10px; font-size:14px; right: 20px; bottom: 20px;'>
            
        <div class='legend-title'>Risk Label Legend</div>
        <ul class='legend-labels'>
            <li><span style='background:green;opacity:0.3;'></span>1</li>
            <li><span style='background:green;opacity:0.5;'></span>2</li>
            <li><span style='background:green;opacity:0.8;'></span>3</li>
            <li><span style='background:yellow;opacity:0.3;'></span>4</li>
            <li><span style='background:blue;opacity:0.3;'></span>5</li>
            <li><span style='background:blue;opacity:0.5;'></span>6</li>
            <li><span style='background:orange;opacity:0.5;'></span>7</li>
            <li><span style='background:red;opacity:0.3;'></span>8</li>
            <li><span style='background:red;opacity:0.5;'></span>9</li>
            <li><span style='background:red;opacity:0.7;'></span>10</li>

        </ul>
        </div>
        
        </body>
        </html>

        <style type='text/css'>
        .maplegend .legend-title {
            text-align: left;
            margin-bottom: 5px;
            font-weight: bold;
            font-size: 90%;
            }
        .maplegend .legend-scale ul {
            margin: 0;
            margin-bottom: 5px;
            padding: 0;
            float: left;
            list-style: none;
            }
        .maplegend .legend-scale ul li {
            font-size: 80%;
            list-style: none;
            margin-left: 0;
            line-height: 18px;
            margin-bottom: 2px;
            }
        .maplegend ul.legend-labels li span {
            display: block;
            float: left;
            height: 16px;
            width: 30px;
            margin-right: 5px;
            margin-left: 0;
            border: 1px solid #999;
            }
        .maplegend .legend-source {
            font-size: 80%;
            color: #777;
            clear: both;
            }
        .maplegend a {
            color: #777;
            }
        </style>
        {% endmacro %}"""

        macro = MacroElement()
        macro._template = Template(template)

        map.get_root().add_child(macro)


        return map


    def AnnualRisk_map(self):
        
        """Create a visualization of predicted annual flood risk to properties
        on a scale of low, medium, high."""
        
        loc = 'Predicted Annual Flood risk to property'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)

        AnnualRisk = Preprocess.Annual_risk_preprocess(self)
        import branca.colormap as cm
        cs = AnnualRisk["class"]

        data = zip(AnnualRisk["lat"],AnnualRisk["long"],cs)

        loc = 'Predicted Annual Flood risk to property'
        title_html = '''
                    <h3 align="center" style="font-size:16px"><b>{}</b></h3>
                    '''.format(loc)

        map = folium.Map([51.4,0.], zoom_start=8,zoom_end =15)

        map.get_root().html.add_child(folium.Element(title_html))

        for la,lo,c in data:
            radius = (c+1)*500
            folium.Circle(location=(la,lo),radius=radius,popup=c,color=False,
            fill_color = colorCall_price(c),fill=True,fill_opacity=0.5).add_to(map)

        template = """
        {% macro html(this, kwargs) %}
        
        <div id='maplegend' class='maplegend' 
            style='position: absolute; z-index:9999; border:2px solid grey; background-color:rgba(255, 255, 255, 0.8);
            border-radius:6px; padding: 10px; font-size:14px; right: 20px; bottom: 20px;'>
            
        <div class='legend-title'>Risk Label Legend</div>
        <ul class='legend-labels'>
            <li><span style='background:green;opacity:0.7;'></span> Low flood risk</li>
            <li><span style='background:blue;opacity:0.7;'></span>Medium flood risk</li>
            <li><span style='background:red;opacity:0.7;'></span>High flood risk</li>
            
        </ul>
        </div>
        
        </body>
        </html>

        <style type='text/css'>
        .maplegend .legend-title {
            text-align: left;
            margin-bottom: 5px;
            font-weight: bold;
            font-size: 90%;
            }
        .maplegend .legend-scale ul {
            margin: 0;
            margin-bottom: 5px;
            padding: 0;
            float: left;
            list-style: none;
            }
        .maplegend .legend-scale ul li {
            font-size: 80%;
            list-style: none;
            margin-left: 0;
            line-height: 18px;
            margin-bottom: 2px;
            }
        .maplegend ul.legend-labels li span {
            display: block;
            float: left;
            height: 16px;
            width: 30px;
            margin-right: 5px;
            margin-left: 0;
            border: 1px solid #999;
            }
        .maplegend .legend-source {
            font-size: 80%;
            color: #777;
            clear: both;
            }
        .maplegend a {
            color: #777;
            }
        </style>
        {% endmacro %}"""

        macro = MacroElement()
        macro._template = Template(template)

        map.get_root().add_child(macro)
        return map




from numpy import array, asarray, mod, sin, cos, tan, sqrt, arctan2, floor, rad2deg, deg2rad, stack, pi
from scipy.linalg import inv
import numpy as np

__all__ = ['get_easting_northing_from_gps_lat_long',
           'get_gps_lat_long_from_easting_northing']


class Ellipsoid(object):
    """ Data structure for a global ellipsoid. """

    def __init__(self, a, b, F_0):
        self.a = a
        self.b = b
        self.n = (a - b) / (a + b)
        self.e2 = (a ** 2 - b ** 2) / a ** 2
        self.F_0 = F_0


class Datum(Ellipsoid):
    """ Data structure for a global datum. """

    def __init__(self, a, b, F_0, phi_0, lam_0, E_0, N_0, H):
        super().__init__(a, b, F_0)
        self.phi_0 = phi_0
        self.lam_0 = lam_0
        self.E_0 = E_0
        self.N_0 = N_0
        self.H = H
        self.af0 = self.a * self.F_0
        self.bf0 = self.b * self.F_0


def dms2rad(deg, min=0, sec=0):
    """Convert degrees, minutes, seconds to radians.

    Parameters
    ----------
    deg: array_like
        Angle in degrees.
    min: array_like
        (optional) Angle component in minutes.
    sec: array_like
        (optional) Angle component in seconds.

    Returns
    -------
    numpy.ndarray
        Angle in radians.
    """
    deg = asarray(deg)
    return deg2rad(deg + min / 60. + sec / 3600.)


def rad2dms(rad, dms=False):
    """Convert radians to degrees, minutes, seconds.

    Parameters
    ----------

    rad: array_like
        Angle in radians.
    dms: bool
        Use degrees, minutes, seconds format. If False, use decimal degrees.

    Returns
    -------
    numpy.ndarray
        Angle in degrees, minutes, seconds or decimal degrees.
    """

    rad = asarray(rad)
    deg = rad2deg(rad)
    if dms:
        min = 60.0 * mod(deg, 1.0)
        sec = 60.0 * mod(min, 1.0)
        return stack((floor(deg), floor(min), sec.round(4)))
    else:
        return deg


osgb36 = Datum(a=6377563.396,
               b=6356256.910,
               F_0=0.9996012717,
               phi_0=deg2rad(49.0),
               lam_0=deg2rad(-2.),
               E_0=400000,
               N_0=-100000,
               H=24.7)

wgs84 = Ellipsoid(a=6378137,
                  b=6356752.3142,
                  F_0=0.9996)


def lat_long_to_xyz(phi, lam, rads=False, datum=osgb36):
    """Convert input latitude/longitude in a given datum into
    Cartesian (x, y, z) coordinates.

    Parameters
    ----------

    phi: array_like
        Latitude in degrees (if radians=False) or radians (if radians=True).
    lam: array_like
        Longitude in degrees (if radians=False) or radians (if radians=True).
    rads: bool (optional)
        If True, input latitudes and longitudes are in radians.
    datum: Datum (optional)
        Datum to use for conversion.
    """
    if not rads:
        phi = deg2rad(phi)
        lam = deg2rad(lam)

    nu = datum.a * datum.F_0 / sqrt(1 - datum.e2 * sin(phi) ** 2)

    return array(((nu + datum.H) * cos(phi) * cos(lam),
                  (nu + datum.H) * cos(phi) * sin(lam),
                  ((1 - datum.e2) * nu + datum.H) * sin(phi)))


def xyz_to_lat_long(x, y, z, rads=False, datum=osgb36):
    p = sqrt(x ** 2 + y ** 2)

    lam = arctan2(y, x)
    phi = arctan2(z, p * (1 - datum.e2))

    for _ in range(10):
        nu = datum.a * datum.F_0 / sqrt(1 - datum.e2 * sin(phi) ** 2)
        dnu = -datum.a * datum.F_0 * cos(phi) * sin(phi) / (1 - datum.e2 * sin(phi) ** 2) ** 1.5

        f0 = (z + datum.e2 * nu * sin(phi)) / p - tan(phi)
        f1 = datum.e2 * (nu ** cos(phi) + dnu * sin(phi)) / p - 1.0 / cos(phi) ** 2
        phi -= f0 / f1

    if not rads:
        phi = rad2dms(phi)
        lam = rad2dms(lam)

    return phi, lam


def get_easting_northing_from_gps_lat_long(phis, lams, rads=False):
    """ Get OSGB36 easting/northing from GPS latitude and longitude pairs.

    Parameters
    ----------
    phis: float/arraylike
        GPS (i.e. WGS84 datum) latitude value(s)
    lams: float/arrayling
        GPS (i.e. WGS84 datum) longitude value(s).
    rads: bool (optional)
        If true, specifies input is is radians.
    Returns
    -------
    numpy.ndarray
        Easting values (in m)
    numpy.ndarray
        Northing values (in m)
        Examples
    --------
    >>> get_easting_northing_from_gps_lat_long([55.5], [-1.54])
    (array([429157.0]), array([623009]))
    References
    ----------
    Based on the formulas in "A guide to coordinate systems in Great Britain".
    See also https://webapps.bgs.ac.uk/data/webservices/convertForm.cfm
    """

    if not isinstance(phis, list) and not isinstance(phis,np.ndarray):
        phis = [phis]
        lams = [lams]
    assert len(phis) == len(lams)

    res_east = []
    res_north = []
    for i, phi in enumerate(phis):
        lam = lams[i]
        if not rads:
            phi = deg2rad(phi)
            lam = deg2rad(lams[i])
        M = bigM(osgb36.bf0, osgb36.n, osgb36.phi_0, phi)
        nu = osgb36.af0 / (sqrt(1 - (osgb36.e2 * ((sin(phi)) ** 2))))
        rho = (nu * (1 - osgb36.e2)) / (1 - (osgb36.e2 * (sin(phi)) ** 2))
        eta2 = (nu / rho) - 1

        I = M + osgb36.N_0
        II = (nu / 2) * sin(phi) * cos(phi)
        III = (nu / 24) * sin(phi) * cos(phi) ** 3 * (5 - tan(phi) ** 2 + 9 * eta2)
        IIIA = (nu / 720) * sin(phi) * cos(phi) ** 5 * (61 - 58 * tan(phi) ** 2 + tan(phi) ** 4)
        IV = nu * cos(phi)
        V = (nu / 6) * cos(phi) ** 3 * (nu / rho - tan(phi) ** 2)
        VI = (nu / 120) * cos(phi) ** 5 * (
                5 - 18 * tan(phi) ** 2 + tan(phi) ** 4 + 14 * eta2 - 58 * tan(phi) ** 2 * eta2)
        diff = lam - osgb36.lam_0
        N = I + II * diff ** 2 + III * diff ** 4 + IIIA * diff ** 6
        E = osgb36.E_0 + IV * diff + V * diff ** 3 + VI * diff ** 5
        res_east.append(round(E,0))
        res_north.append(round(N,0))

    return res_east, res_north


def bigM(bf0, n, PHI0, PHI):
    """
    Compute meridional arc.
    Input:
     - ellipsoid semi major axis multiplied by central meridian scale factor (bf0) in meters;
     - n (computed from a, b and f0);
     - lat of false origin (PHI0) and initial or final latitude of point (PHI) IN RADIANS.
    """
    M = bf0 * (((1 + n + ((5 / 4) * (n ** 2)) + ((5 / 4) * (n ** 3))) * (PHI - PHI0))
               - (((3 * n) + (3 * (n ** 2)) + ((21 / 8) * (n ** 3))) * (sin(PHI - PHI0)) * (
                cos(PHI + PHI0)))
               + ((((15 / 8) * (n ** 2)) + ((15 / 8) * (n ** 3))) * (sin(2 * (PHI - PHI0))) * (
                cos(2 * (PHI + PHI0))))
               - (((35 / 24) * (n ** 3)) * (sin(3 * (PHI - PHI0))) * (cos(3 * (PHI + PHI0)))))
    return M


def getNewPhi(North, datum):
    PHI1 = ((North - datum.N_0) / datum.af0) + datum.phi_0
    M = bigM(datum.bf0, datum.n, datum.phi_0, PHI1)

    # Calculate new PHI value (PHI2)
    PHI2 = ((North - datum.N_0 - M) / datum.af0) + PHI1

    # Iterate to get final value for InitialLat
    while abs(North - datum.N_0 - M) > 0.00001:
        PHI2 = ((North - datum.N_0 - M) / datum.af0) + PHI1
        M = bigM(datum.bf0, datum.n, datum.phi_0, PHI2)
        PHI1 = PHI2

    return PHI2


def get_gps_lat_long_from_easting_northing(easts: float, norths: float, rads=False, dms=False):
    """ Get OSGB36 easting/northing from GPS latitude and
    longitude pairs.
    Parameters
    ----------
    east: float/arraylike
        OSGB36 easting value(s) (in m).
    north: float/arrayling
        OSGB36 easting value(s) (in m).
    rads: bool (optional)
        If true, specifies ouput is radians.
    dms: bool (optional)
        If true, output is in degrees/minutes/seconds. Incompatible
        with rads option.
    Returns
    -------
    numpy.ndarray
        GPS (i.e. WGS84 datum) latitude value(s).
    numpy.ndarray
        GPS (i.e. WGS84 datum) longitude value(s).
    Examples
    --------
    >>> get_gps_lat_long_from_easting_northing([429157], [623009])
    (array([55.5]), array([-1.540008]))
    References
    ----------
    Based on the formulas in "A guide to coordinate systems in Great Britain".
    See also https://webapps.bgs.ac.uk/data/webservices/convertForm.cfm
    """
    if not isinstance(easts, list) and not isinstance(easts,np.ndarray):
        easts = [easts]
        norths = [norths]
    assert len(easts) == len(norths)
    res_lat = []
    res_lon = []
    for i, east in enumerate(easts):
        north = norths[i]
        Et = east - osgb36.E_0
        PHId = getNewPhi(north, datum=osgb36)
        nu = osgb36.af0 / (sqrt(1 - (osgb36.e2 * ((sin(PHId)) ** 2))))
        rho = (nu * (1 - osgb36.e2)) / (1 - (osgb36.e2 * (sin(PHId)) ** 2))
        eta2 = (nu / rho) - 1

        # Compute Latitude
        VII = (tan(PHId)) / (2 * rho * nu)
        VIII = ((tan(PHId)) / (24 * rho * (nu ** 3))) * (
                5 + (3 * ((tan(PHId)) ** 2)) + eta2 - (9 * eta2 * ((tan(PHId)) ** 2)))
        IX = ((tan(PHId)) / (720 * rho * (nu ** 5))) * (
                61 + (90 * ((tan(PHId)) ** 2)) + (45 * ((tan(PHId)) ** 4)))

        X = ((cos(PHId)) ** -1) / nu
        XI = (((cos(PHId)) ** -1) / (6 * (nu ** 3))) * ((nu / rho) + (2 * ((tan(PHId)) ** 2)))
        XII = (((cos(PHId)) ** -1) / (120 * (nu ** 5))) * (
                5 + (28 * ((tan(PHId)) ** 2)) + (24 * ((tan(PHId)) ** 4)))
        XIIA = (((cos(PHId)) ** -1) / (5040 * (nu ** 7))) * (
                61 + (662 * ((tan(PHId)) ** 2)) + (1320 * ((tan(PHId)) ** 4)) + (
                720 * ((tan(PHId)) ** 6)))
        E_N_Lat = PHId - ((Et ** 2) * VII) + ((Et ** 4) * VIII) - ((Et ** 6) * IX)
        E_N_Long = osgb36.lam_0 + (Et * X) - ((Et ** 3) * XI) + ((Et ** 5) * XII) - ((Et ** 7) * XIIA)
        if not rads:
            E_N_Lat = rad2dms(E_N_Lat, dms=dms)
            E_N_Long = rad2dms(E_N_Long, dms=dms)
        res_lat.append(E_N_Lat)
        res_lon.append(E_N_Long)
    return res_lat, res_lon


class HelmertTransform(object):
    """Callable class to perform a Helmert transform."""

    def __init__(self, s, rx, ry, rz, T):
        self.T = T.reshape((3, 1))

        self.M = array([[1 + s, -rz, ry],
                        [rz, 1 + s, -rx],
                        [-ry, rx, 1 + s]])

    def __call__(self, X):
        X = X.reshape((3, -1))
        return self.T + self.M @ X


class HelmertInverseTransform(object):
    """Callable class to perform the inverse of a Helmert transform."""

    def __init__(self, s, rx, ry, rz, T):
        self.T = T.reshape((3, 1))

        self.M = inv(array([[1 + s, -rz, ry],
                            [rz, 1 + s, -rx],
                            [-ry, rx, 1 + s]]))

    def __call__(self, X):
        X = X.reshape((3, -1))
        return self.M @ (X - self.T)


OSGB36transform = HelmertTransform(20.4894e-6,
                                   -dms2rad(0, 0, 0.1502),
                                   -dms2rad(0, 0, 0.2470),
                                   -dms2rad(0, 0, 0.8421),
                                   array([-446.448, 125.157, -542.060]))

WGS84transform = HelmertInverseTransform(20.4894e-6,
                                         -dms2rad(0, 0, 0.1502),
                                         -dms2rad(0, 0, 0.2470),
                                         -dms2rad(0, 0, 0.8421),
                                         array([-446.448, 125.157, -542.060]))


def WGS84toOSGB36(phi, lam, rads=False):
    """Convert WGS84 latitude/longitude to OSGB36 latitude/longitude.

    Parameters
    ----------
    phi : array_like or float
        Latitude in degrees or radians on WGS84 datum.
    lam : array_like or float
        Longitude in degrees or radians on WGS84 datum.
    rads : bool, optional
        If True, phi and lam are in radians. If False, phi and lam are in degrees.

    Returns
    -------
    tuple of numpy.ndarrays
        Latitude and longitude on OSGB36 datum in degrees or radians.
    """
    xyz = OSGB36transform(lat_long_to_xyz(asarray(phi), asarray(lam),
                                          rads=rads, datum=wgs84))
    return xyz_to_lat_long(*xyz, rads=rads, datum=osgb36)


def OSGB36toWGS84(phi, lam, rads=False):
    """Convert OSGB36 latitude/longitude to WGS84 latitude/longitude.

    Parameters
    ----------
    phi : array_like or float
        Latitude in degrees or radians on OSGB36 datum.
    lam : array_like or float
        Longitude in degrees or radians on OSGB36 datum.
    rads : bool, optional
        If True, phi and lam are in radians. If False, phi and lam are in degrees.

    Returns
    -------
    tuple of numpy.ndarrays
        Latitude and longitude on WGS84 datum in degrees or radians.
    """
    xyz = WGS84transform(lat_long_to_xyz(asarray(phi), asarray(lam),
                                         rads=rads, datum=osgb36))
    return xyz_to_lat_long(*xyz, rads=rads, datum=wgs84)


if __name__ == "__main__":
    print(get_easting_northing_from_gps_lat_long([55.5], [-1.54]))



"""Interactions with rainfall and river data."""

import numpy as np
import pandas as pd

__all__ = ["get_station_data_from_csv"]


def get_station_data_from_csv(filename, station_reference):
    """Return readings for a specified recording station from .csv file.

    Parameters
    ----------

    filename: str
        filename to read
    station_reference
        station_reference to return.

    >>> data = get_station_data_from_csv('resources/wet_day.csv')
    """
    frame = pd.read_csv(filename)
    frame = frame.loc[frame.stationReference == station_reference]

    return pd.to_numeric(frame.value.values)

def get_live_station_data(station_reference):
    
    """Return readings for a specified recording station from live API as a DataFrame.

    Parameters
    ----------

    station_reference: str or list/ndarray of str 
        station_reference to return.

    >>> data = get_live_station_data('0184TH')
    """
    
    assert isinstance(station_reference, str) or isinstance(station_reference, np.ndarray) or isinstance(station_reference, list), 'Station reference must be passed as a string, list or np.ndarray'

    if isinstance(station_reference, str):
        station_reference = [station_reference]
    
    value_list = []
    dateTime_list = []
    lat_list = []
    long_list = []
    parameter_list = []

    for j in station_reference:
        try:
            df_live = pd.read_json(f'https://environment.data.gov.uk/flood-monitoring/id/stations/{j}')
        except:
            print(f' station reference {j} does not exist live')
            continue
            

        station_info = df_live.loc['measures', 'items']
        
        if isinstance(station_info, list):
            station_info = station_info[0]
        
        assert  isinstance(station_info, dict), f'{j} has a different format'
        
        if not isinstance(station_info.get('latestReading'), dict):
            value =  'Unavailable'
            parameter =  'Unavailable'
            lat = 'Unavailable'
            long = 'Unavailable'
            
        else:
            value = station_info.get('latestReading').get('value')
            value_list.append(value)
        
            dateTime = station_info.get('latestReading').get('dateTime')
            dateTime_list.append(dateTime)

            parameter = station_info.get('parameter')
            parameter_list.append(parameter)
            
        if 'lat' in df_live.index:
            lat = df_live.loc['lat', 'items']
            lat_list.append(lat)
        else:
            lat =  'Unavailable'
            
        if 'long' in df_live.index:  
            long = df_live.loc['long', 'items']
            long_list.append(long)
        else:
            long = 'Unavailable'
        

    station_data = pd.DataFrame(data= dateTime_list, columns=['dateTime'])
    station_data ['latitude'] = lat_list
    station_data ['longitude'] = long_list
    station_data ['parameter'] = parameter_list
    station_data ['value'] = value_list
   

    return station_data
    
    




"""Analysis tools."""


import os
import math

import numpy as np
import pandas as pd

import matplotlib
import matplotlib.pyplot as plt

__all__ = ['plot_risk_map']

def plot_postcode_density(postcode_file=(os.path.dirname(__file__)
                                      +'/resources/postcodes_unlabelled.csv'),
                          coordinate=['easting','northing'], dx=1000):
    """Plot a postcode density map from a postcode file."""

    
    pdb = pd.read_csv(postcode_file)

    

    bbox = (pdb[coordinate[0]].min()-0.5*dx, pdb[coordinate[0]].max()+0.5*dx,
            pdb[coordinate[1]].min()-0.5*dx, pdb[coordinate[1]].max()+0.5*dx)

    
    nx = (math.ceil((bbox[1] - bbox[0])/dx),
          math.ceil((bbox[3] - bbox[2])/dx))

    x = np.linspace(bbox[0]+0.5 * dx, bbox[0] + (nx[0] - 0.5) * dx, nx[0])
    y = np.linspace(bbox[2]+0.5 * dx, bbox[2] + (nx[1] - 0.5) * dx, nx[1])

    X, Y = np.meshgrid(x, y)

    Z = np.zeros(nx, int)

    for x, y in pdb[coordinate].values:
        Z[math.floor((x-bbox[0])/dx), math.floor((y-bbox[2])/dx)] += 1

    plt.pcolormesh(X, Y, np.where(Z>0, Z, np.nan).T,
                   norm=matplotlib.colors.LogNorm())
    plt.axis('equal')
    plt.colorbar()
    plt.show()
    

def plot_risk_map(risk_data, coordinate=['easting','northing'], dx=1000):
    """Plot a risk map."""

    bbox = (risk_data[coordinate[0]].min()-0.5*dx, risk_data[coordinate[0]].max()+0.5*dx,
            risk_data[coordinate[1]].min()-0.5*dx, risk_data[coordinate[1]].max()+0.5*dx)

    
    nx = (math.ceil((bbox[1] - bbox[0])/dx),
          math.ceil((bbox[3] - bbox[2])/dx))

    x = np.linspace(bbox[0]+0.5 * dx, bbox[0] + (nx[0] - 0.5) * dx, nx[0])
    y = np.linspace(bbox[2]+0.5 * dx, bbox[2] + (nx[1] - 0.5) * dx, nx[1])

    X, Y = np.meshgrid(x, y)

    Z = np.zeros(nx, int)

    for x, y, val in risk_data[coordinate+'risk'].values:
        Z[math.floor((x-bbox[0])/dx), math.floor((y-bbox[2])/dx)] += val

    plt.pcolormesh(X, Y, np.where(Z>0, Z, np.nan).T,
                   norm=matplotlib.colors.LogNorm())
    plt.axis('equal')
    plt.colorbar()
    
        
        
if __name__=="__main__":
    plot_postcode_density("/Users/ms4922/Documents/ads-deluge-Nene/flood_tool/resources/postcodes_sampled.csv")
    



import folium

all = ['plot_circle']


def plot_circle(lat, lon, radius=2000, map=None, **kwargs):
    """
    Plot a circle on a map (creating a new folium map instance if necessary).

    Parameters
    ----------

    lat: float
        latitude of circle to plot (degrees)
    lon: float
        longitude of circle to plot (degres)
    radius: float
        radius of circle to plot (m)
    map: folium.Map
        existing map object

    Returns
    -------

    Folium map object

    Examples
    --------

    >>> import folium
    >>> plot_circle(52.79, -2.95, 1e3, map=None)
    """

    if not map:
        map = folium.Map(location=(lat, lon), control_scale=True)


    
    folium.Circle(location=[lat, lon], radius=radius, fill=True,
                  fillOpacity=0.6, **kwargs).add_to(map)

    return map




"""Example module in template package."""

import os

import numpy as np
import pandas as pd

from .geo import *
from sklearn.preprocessing import LabelEncoder
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from imblearn.over_sampling import SMOTE
from imblearn.over_sampling import RandomOverSampler

import joblib

__all__ = ['Tool']


class Tool(object):
    """Class to interact with a postcode database file."""

    def __init__(self, household_file='', postcode_file='', sample_labels=''):
        """
        Parameters
        ----------

        full_postcode_file : str, optional
            Filename of a .csv file containing geographic location
            data for postcodes.

        household_file : str, optional
            Filename of a .csv file containing information on households
            by postcode.
        """

        if postcode_file == '':
            full_postcode_file = os.sep.join((os.path.dirname(__file__),
                                              'resources',
                                              'postcodes_unlabelled.csv'))

        if household_file == '':
            household_file = os.sep.join((os.path.dirname(__file__),
                                          'resources',
                                          'households_per_sector.csv'))

        self.postcodedb = pd.read_csv(full_postcode_file)
        self.householddb = pd.read_csv(household_file)
        self.model_dic = {
            "flood_label_model": None,
            "price_model_for_lower_price": None,
            "price_model_for_all": None,
            "authority_model": None,
            "rank_classifier": None,
            "price_model_for_higher_price":None
        }

        self.preprocess(sample_labels)

    def preprocess(self, labelled_samples=""):
        if labelled_samples == '':
            labelled_samples = os.sep.join((os.path.dirname(__file__),
                                            'resources',
                                            'postcodes_sampled.csv'))

        self.train_data = pd.read_csv(labelled_samples)
        flood_label_location_data = self.train_data[['easting', 'northing', 'riskLabel']]
        y = flood_label_location_data[['riskLabel']]
        X = flood_label_location_data[['easting', 'northing']]

        ros = RandomOverSampler(random_state=0)
        self.x_over, self.y_over = ros.fit_resample(X, y)
        # get house price
        house_price_data = self.train_data[['easting', 'northing', 'medianPrice']]
        house_price_data['rank'] = pd.cut(x=house_price_data['medianPrice'],
                                          bins=[house_price_data['medianPrice'].min() - 1,
                                                1000000,
                                                house_price_data['medianPrice'].max() + 1],
                                          labels=['low', 'high'])

        self.X_rank_price = house_price_data[['easting', 'northing']]
        self.y_rank_price = house_price_data[['rank']]

        high_price_data = house_price_data[house_price_data['rank'] == 'high'].drop(columns='rank')
        low_price_data = house_price_data[house_price_data['rank'] == 'low'].drop(columns='rank')

        # high low model's data
        self.X_high_price = high_price_data[['easting', 'northing']]

        self.y_high_price = high_price_data[['medianPrice']]

        self.X_low_price = low_price_data[['easting', 'northing']]

        self.y_low_price = low_price_data[['medianPrice']]

        self.X_price_total = self.train_data[['easting', 'northing']]

        self.y_price_total = self.train_data[['medianPrice']]
        # get authority

        authority_train_data = self.train_data[['easting', 'northing', 'localAuthority']]

        self.X_authority = authority_train_data[['easting', 'northing']]

        self.y_authority = authority_train_data[['localAuthority']]

        self.authority_lencoder = LabelEncoder().fit(self.y_authority)

    def train(self, labelled_samples=''):
        """Train the model using a labelled set of samples.
        
        Parameters
        ----------
        
        labelled_samples : str, optional
            Filename of a .csv file containing a labelled set of samples.
        """
        self.model_dic = {}
        flood_label_model = RandomForestClassifier().fit(self.x_over, self.y_over)

        self.model_dic['flood_label_model'] = flood_label_model

        # get house price_high_low
        ##rank predict
        rank_classifier = KNeighborsClassifier().fit(self.X_rank_price, self.y_rank_price)

        self.model_dic['rank_classifier'] = rank_classifier

        price_model_lower = KNeighborsRegressor(n_neighbors=8, weights='distance').fit(self.X_low_price,
                                                                                       self.y_low_price)

        self.model_dic['price_model_for_lower_price'] = price_model_lower

        price_model_high = KNeighborsRegressor(n_neighbors=8, weights='distance').fit(self.X_high_price,
                                                                                      self.y_high_price)

        self.model_dic['price_model_for_higher_price'] = price_model_high

        # get house price_all
        price_model_total = KNeighborsRegressor(algorithm='brute', n_neighbors=9, weights='distance')

        price_model_total.fit(self.X_price_total, self.y_price_total)

        self.model_dic['price_model_for_all'] = price_model_total

        y_authority = pd.Series(self.authority_lencoder.transform(self.y_authority))

        smote = SMOTE(random_state=1)

        X_smote, y_smote = smote.fit_resample(self.X_authority, y_authority)

        authority_model = RandomForestClassifier().fit(X_smote, y_smote)

        self.model_dic['authority_model'] = authority_model
        weights_path = os.path.join(os.path.dirname(__file__), "weights")
        if not os.path.exists(weights_path):
            os.mkdir(weights_path)
        for name, model in self.model_dic.items():
            joblib.dump(model, os.path.join(weights_path, name + ".pkl"))

    def load(self):
        weights_path = os.path.join(os.path.dirname(__file__), "weights")
        if os.path.exists(os.path.join(weights_path, list(self.model_dic.keys())[0] + '.pkl')):
            for name in self.model_dic.keys():
                self.model_dic[name] = joblib.load(os.path.join(weights_path, name + ".pkl"))
        else:
            raise BaseException("Weights file don't exist!")

    def get_useful_features(self, postcodes,
                            features=['sector', 'easting', 'northing', 'localAuthority', 'altitude', 'soilType']):
        if isinstance(postcodes, pd.Series):
            postcodes = postcodes.to_list()
        if not isinstance(postcodes, list) and not isinstance(postcodes, np.ndarray):
            postcodes = [postcodes]
        frame = self.postcodedb.copy()
        frame = frame.set_index('postcode')

        return frame.loc[postcodes, features]

    def get_easting_northing(self, postcodes):
        """Get a frame of OS eastings and northings from a collection
        of input postcodes.

        Parameters
        ----------

        postcodes: sequence of strs
            Sequence of postcodes.

        Returns
        -------

        pandas.DataFrame
            DataFrame containing only OSGB36 easthing and northing indexed
            by the input postcodes. Invalid postcodes (i.e. not in the
            input unlabelled postcodes file) return as NaN.
         """
        if isinstance(postcodes, pd.Series):
            postcodes = postcodes.to_list()
        if not isinstance(postcodes, list) and not isinstance(postcodes, np.ndarray):
            postcodes = [postcodes]


        frame = self.postcodedb.copy()
        if not sum(frame["postcode"].isin(postcodes)):
            return None

        frame = frame.set_index('postcode')
        # Convert to Capital
        for i, postcode in enumerate(postcodes):
            postcodes[i] = postcode.upper()

        return frame.loc[postcodes, ['easting', 'northing']]

    def get_lat_long(self, postcodes):
        """Get a frame containing GPS latitude and longitude information for a
        collection of postcodes.

        Parameters
        ----------

        postcodes: sequence of strs
            Sequence of postcodes.

        Returns
        -------

        pandas.DataFrame
            DataFrame containing only WGS84 latitude and longitude pairs for
            the input postcodes. Invalid postcodes (i.e. not in the
            input unlabelled postcodes file) return as NAN.
        """
        if isinstance(postcodes, pd.Series):
            postcodes = postcodes.to_list()
        if not isinstance(postcodes, list) and not isinstance(postcodes, np.ndarray):
            postcodes = [postcodes]

        if not sum(self.postcodedb["postcode"].isin(postcodes)):
            return None
        es = self.get_easting_northing(postcodes)
        lat, lon = get_gps_lat_long_from_easting_northing(es["easting"].to_list(), es["northing"].to_list())

        return pd.DataFrame(np.concatenate([[lat], [lon]], axis=0).T, columns=["latitude", "longitude"], index=es.index)

    @staticmethod
    def get_flood_class_from_postcodes_methods():
        """
        Get a dictionary of available flood probablity classification methods
        for postcodes.

        Returns
        -------

        dict
            Dictionary mapping classification method names (which have
             no inate meaning) on to an identifier to be passed to the
             get_flood_class_from_postcode method.
        """
        flood_class_method = {'all_zero_risk': 0, 'RandomForestClassifier': 1}
        return flood_class_method

    def get_flood_class_from_postcodes(self, postcodes, method=1):
        """
        Generate series predicting flood probability classification
        for a collection of poscodes.

        Parameters
        ----------

        postcodes : sequence of strs
            Sequence of postcodes.
        method : int (optional)
            optionally specify (via a value in
            get_flood_class_from_postcodes_methods) the classification
            method to be used.

        Returns
        -------

        pandas.Series
            Series of flood risk classification labels indexed by postcodes.
        """
        if isinstance(postcodes, pd.Series):
            postcodes = postcodes.to_list()
        if not isinstance(postcodes, list) and not isinstance(postcodes, np.ndarray):
            postcodes = [postcodes]

        if not sum(self.postcodedb["postcode"].isin(postcodes)):
            return None

        assert method in [0, 1], 'No such a method'
        if method == 0:
            return pd.Series(data=np.ones(len(postcodes), int),
                             index=np.asarray(postcodes),
                             name='riskLabel')
        elif method == 1:
            OSBG36_cord = self.get_easting_northing(postcodes)
            eastings = OSBG36_cord['easting'].to_list()
            northings = OSBG36_cord['northing'].to_list()
            OS_RLPredic = self.get_flood_class_from_OSGB36_locations(eastings, northings, method=method)

            PC_RLPrediction_Series = pd.Series(data=OS_RLPredic.values,
                                               index=[postcode for postcode in postcodes],
                                               name='riskLabel')

            return PC_RLPrediction_Series
        else:
            raise NotImplementedError

    @staticmethod
    def get_flood_class_from_locations_methods():
        """
        Get a dictionary of available flood probablity classification methods
        for locations.

        Returns
        -------

        dict
            Dictionary mapping classification method names (which have
             no inate meaning) on to an identifier to be passed to the
             get_flood_class_from_OSGB36_locations and
             get_flood_class_from_OSGB36_locations method.
        """

        flood_class_method = {'Do Nothing': 0, 'RandomForestClassifier': 1}
        return flood_class_method

    def get_flood_class_from_OSGB36_locations(self, eastings, northings, method=1):
        """
        Generate series predicting flood probability classification
        for a collection of OSGB36_locations.

        Parameters
        ----------

        eastings : sequence of floats
            Sequence of OSGB36 eastings.
        northings : sequence of floats
            Sequence of OSGB36 northings.
        method : int (optional)
            optionally specify (via a value in
            self.get_flood_class_from_locations_methods) the classification
            method to be used.

        Returns
        -------

        pandas.Series
            Series of flood risk classification labels indexed by locations.
        """
        if isinstance(eastings, pd.Series):
            eastings = eastings.to_list()
            northings = northings.to_list()
        if not isinstance(eastings, list) and not isinstance(eastings, np.ndarray):
            eastings = [eastings]
            northings = [northings]
        assert len(northings) == len(eastings)
        assert method in [0, 1], 'No such a method'

        if method == 0:
            return pd.Series(data=np.ones(len(eastings), int),
                             index=[(est, nth) for est, nth in
                                    zip(eastings, northings)],
                             name='riskLabel')
        elif method == 1:
            cor_list = []
            for est, nor in zip(eastings, northings):
                cor_list.append([est, nor])

            model = self.model_dic['flood_label_model']

            RLPrediction = model.predict(pd.DataFrame(cor_list))

            RLPrediction_Series = pd.Series(data=RLPrediction,
                                            index=[(est, nth) for est, nth in
                                                   zip(eastings, northings)],
                                            name='riskLabel')
            return RLPrediction_Series
        else:
            raise NotImplementedError

    def get_flood_class_from_WGS84_locations(self, latitudes, longitudes, rad=False, method=0):
        """
        Generate series predicting flood probability classification
        for a collection of WGS84 datum locations.

        Parameters
        ----------

        longitudes : sequence of floats
            Sequence of WGS84 longitudes.
        latitudes : sequence of floats
            Sequence of WGS84 latitudes.
        method : int (optional)
            optionally specify (via a value in
            self.get_flood_class_from_locations_methods) the classification
            method to be used.

        Returns
        -------

        pandas.Series
            Series of flood risk classification labels indexed by locations.
        """

        assert method in [0, 1], 'No such a method'
        if isinstance(latitudes, pd.Series):
            latitudes = latitudes.to_list()
            longitudes = longitudes.to_list()
        if not isinstance(latitudes, list) and not isinstance(latitudes, np.ndarray):
            latitudes = [latitudes]
            longitudes = [longitudes]
        assert len(latitudes) == len(longitudes)
        if method == 0:
            return pd.Series(data=np.ones(len(longitudes), int),
                             index=[(lng, lat) for lng, lat in
                                    zip(longitudes, latitudes)],
                             name='riskLabel')
        else:
            # convert from WGS84 to OSGB36
            eastings, northings = get_easting_northing_from_gps_lat_long(latitudes, longitudes, rads=rad)
            OS_RLPredic = self.get_flood_class_from_OSGB36_locations(eastings, northings, method)

        WG_RLPrediction_Series = pd.Series(data=OS_RLPredic.values,
                                           index=[(lat, lot) for lat, lot in
                                                  zip(latitudes, longitudes)],
                                           name='riskLabel')

        return WG_RLPrediction_Series

    @staticmethod
    def get_house_price_methods():
        """
        Get a dictionary of available flood house price regression methods.

        Returns
        -------

        dict
            Dictionary mapping regression method names (which have
             no inate meaning) on to an identifier to be passed to the
             get_median_house_price_estimate method.
        """
        return {'all_england_median': 0, 'KNeighborsRegressor_total_value': 1,
                'KNeighborsRegressor_split_price_area': 2}

    def get_median_house_price_estimate(self, postcodes, method=2):
        """
        Generate series predicting median house price for a collection
        of poscodes.

        Parameters
        ----------

        postcodes : sequence of strs
            Sequence of postcodes.
        method : int (optional)
            optionally specify (via a value in
            self.get_house_price_methods) the regression
            method to be used.

        Returns
        -------

        pandas.Series
            Series of median house price estimates indexed by postcodes.
        """
        if isinstance(postcodes, pd.Series):
            postcodes = postcodes.to_list()
        if not isinstance(postcodes, list) and not isinstance(postcodes, np.ndarray):
            postcodes = [postcodes]

        if not sum(self.postcodedb["postcode"].isin(postcodes)):
            return None

        if method == 0:
            return pd.Series(data=np.full(len(postcodes), 245000.0),
                             index=np.asarray(postcodes),
                             name='medianPrice')

        elif method == 1:

            model = self.model_dic['price_model_for_all']
            X_pre = self.get_useful_features(postcodes, ['easting', 'northing'])

            return pd.Series(index=np.asarray(postcodes), data=np.asarray(model.predict(X_pre)).flatten(),
                             name='medianPrice')
        elif method == 2:

            X_pre = self.get_useful_features(postcodes, ['easting', 'northing'])

            # get rank predict model
            rank_model = self.model_dic['rank_classifier']
            X_pre['rank'] = rank_model.predict(X_pre)

            # get differnet price set
            X_high = X_pre[X_pre['rank'] == 'high'].drop(columns='rank')
            X_low = X_pre[X_pre['rank'] == 'low'].drop(columns='rank')
            # get trained model
            low_model = self.model_dic['price_model_for_lower_price']
            high_model = self.model_dic['price_model_for_higher_price']
            if len(X_low) and len(X_high):
                low_result = pd.Series(index=X_low.index, data=np.asarray(low_model.predict(X_low).flatten()),
                                       name='medianPrice')
                high_result = pd.Series(index=X_high.index, data=np.asarray(high_model.predict(X_high).flatten()),
                                        name='medianPrice')

                return pd.concat([low_result, high_result])
            elif len(X_high):
                return pd.Series(index=X_high.index, data=np.asarray(high_model.predict(X_high).flatten()),
                                 name='medianPrice')
            elif len(X_low):
                return pd.Series(index=X_low.index, data=np.asarray(low_model.predict(X_low).flatten()),
                                 name='medianPrice')

        else:
            raise NotImplementedError

    @staticmethod
    def get_local_authority_methods():
        """
        Get a dictionary of available local authorithy classification methods.

        Returns
        -------

        dict
            Dictionary mapping regression method names (which have
             no inate meaning) on to an identifier to be passed to the
             get_altitude_estimate method.
        """

        class_method = {'Do Nothing': 0, 'RandomForestClassifier': 1}
        return class_method

    def get_local_authority_estimate(self, eastings, northings, method=1):
        """
        Generate series predicting local authorities in n for a sequence
        of OSGB36 locations.

        Parameters
        ----------

        eastingss : sequence of floats
            Sequence of OSGB36 eastings.
        northings : sequence of floats
            Sequence of OSGB36 northings.
        method : int (optional)
            optionally specify (via a value in
            self.get_altitude_methods) the regression
            method to be used.

        Returns
        -------

        pandas.Series
            Series of LocalAuthority indexed by OSGB36 locations.
        """
        assert method in [0, 1], 'No such a method'
        if isinstance(eastings, pd.Series):
            eastings = eastings.to_list()
            northings = northings.to_list()
        if not isinstance(eastings, list) and not isinstance(eastings, np.ndarray):
            eastings = [eastings]
            northings = [northings]
        assert len(northings) == len(eastings)
        if method == 0:
            return pd.Series(data=np.full(len(eastings), 'Unknown'),
                             index=[(est, nth) for est, nth in
                                    zip(eastings, northings)],
                             name='localAuthority')
        elif method == 1:
            model = self.model_dic['authority_model']
            cor_list = []
            for est, nor in zip(eastings, northings):
                cor_list.append([est, nor])

            LAPrediction = model.predict(pd.DataFrame(cor_list))
            LAPrediction = self.authority_lencoder.inverse_transform(LAPrediction)
            LAPrediction_Series = pd.Series(data=LAPrediction,
                                            index=[(est, nth) for est, nth in
                                                   zip(eastings, northings)],
                                            name='localAuthority')
            return LAPrediction_Series
        else:
            raise NotImplementedError

    def get_total_value(self, postal_data):
        """
        Return a series of estimates of the total property values
        of a sequence of postcode units or postcode sectors.


        Parameters
        ----------

        postal_data : sequence of strs
            Sequence of postcode units or postcodesectors


        Returns
        -------

        pandas.Series
            Series of total property value estimates indexed by locations.
        """

        if isinstance(postal_data, pd.Series):
            postal_data = postal_data.to_list()
        if not isinstance(postal_data, list) and not isinstance(postal_data, np.ndarray):
            postal_data = [postal_data]

        unlabel_postcode_file = os.sep.join((os.path.dirname(__file__),
                                             'resources',
                                             'unlabeledsampleWithHouseholdAndNearestStation.csv'))
        unlabeled_df = pd.read_csv(unlabel_postcode_file)
        length = postal_data[0].split(" ")
        postal_data_copy = postal_data.copy()
        if len(length[1]) != 3:
            postal_data = unlabeled_df[unlabeled_df['sector'].isin(postal_data)]['postcode'].tolist()
        median_price = self.get_median_house_price_estimate(postal_data)
        households = unlabeled_df[unlabeled_df['postcode'].isin(postal_data)]['households']
        households.index = postal_data
        num_units = unlabeled_df[unlabeled_df['postcode'].isin(postal_data)]['number of postcode units']
        num_units.index = postal_data
        total = median_price * households / num_units
        total.index = postal_data
        total.name = 'total property value'
        return total

    def get_annual_flood_risk(self, postcodes, risk_labels=None):
        """
        Return a series of estimates of the total property values of a
        collection of postcodes.

        Risk is defined here as a damage coefficient multiplied by the
        value under threat multiplied by the probability of an event.

        Parameterszz
        ----------

        postcodes : sequence of strs
            Sequence of postcodes.
        risk_labels: pandas.Series (optional)
            Series containing flood risk classifiers, as
            predicted by get_flood_class_from_postcodes.

        Returns
        -------

        pandas.Series
            Series of total annual flood risk estimates indexed by locations.
        """
        if isinstance(postcodes, pd.Series):
            postcodes = postcodes.to_list()
        if not isinstance(postcodes, list) and not isinstance(postcodes, np.ndarray):
            postcodes = [postcodes]

        if not sum(self.postcodedb["postcode"].isin(postcodes)):
            return None

        risk_labels = risk_labels or self.get_flood_class_from_postcodes(postcodes)

        cost = self.get_total_value(postcodes)

        damage_coefficient = 0.05

        risk_proba = [0.0001, 0.0005, 0.001, 0.005, 0.01, 0.015, 0.02, 0.03, 0.04, 0.05]
        risk_proba = np.array(risk_proba)

        risk_df = pd.DataFrame(data=risk_proba.T, columns=['risk_proba'], index=list(range(1, 11)))
        risk_proba = risk_df.loc[risk_labels.to_list()]["risk_proba"]
        risk_proba.index = postcodes

        annual_flood_risk = damage_coefficient * cost * risk_proba

        annual_flood_risk.name = 'Annual Flood Risk'
        return annual_flood_risk



#!/usr/bin/env python

from distutils.core import setup, Extension
from os import path
this_directory = path.abspath(path.dirname(__file__))
import pypandoc
#long_descriptions = pypandoc.convert_file('README.md', 'rst')

setup(
      name='FloodToolNene',
      version='1.0',
      description='Flood Risk Analysis Tool',
      author='ADS project Team Nene',
      url = 'https://github.com/ese-msc-2022/ads-deluge-Nene',
      #long_description = long_descriptions,
      packages=['flood_tool']
      )
